(* $Id: dbll_adabas.ml,v 1.2 1999/01/28 02:06:36 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(* TODO:
 * - choose adabas mode, timeouts, isolation level
 * - should avoid to fill the SQLDA structure for every statement request
 *   again
 *)

open Db_types


(**********************************************************************)
(*                                                                    *)
(* External functions                                                 *)
(*                                                                    *)
(**********************************************************************)

type sqlda;;
   (* an SQL descriptor area. This is a C structure used to pass host
    * variables to and from the DBMS
    *)


(*** statements ***)

external sql_connect : int -> string -> string -> string -> string -> 
  int -> int -> int -> unit
    = "dbll_adabas_connect_bytecode" "dbll_adabas_connect_native"
external sql_transaction : int -> int -> unit
    = "dbll_adabas_transaction"
external sql_prepare : int -> int -> string -> unit
    = "dbll_adabas_prepare"
external sql_describe : int -> int -> sqlda
    = "dbll_adabas_describe"
external sql_describe_fetch_cursor : int -> int -> sqlda
    = "dbll_adabas_describe_fetch_cursor"
external sql_execute : int -> int -> sqlda -> unit
    = "dbll_adabas_execute"
external sql_declare_cursor : int -> int -> unit
    = "dbll_adabas_declare_cursor"
external sql_open_cursor : int -> int -> sqlda -> unit
    = "dbll_adabas_open_cursor"
external sql_fetch_cursor : int -> int -> sqlda -> unit
    = "dbll_adabas_fetch_cursor"
external sql_close_cursor : int -> int -> unit
    = "dbll_adabas_close_cursor"

(*** SQLCA structure ***)

(* SQLCA = condition area *)

external sqlca_code : unit -> int
    = "dbll_adabas_sqlca_code"

external sqlca_message : unit -> string
    = "dbll_adabas_sqlca_message"


(*** SQLDA structure ***)

external sqlda_length : sqlda -> int
    = "dbll_adabas_length"
external sqlda_get : sqlda -> int -> (int * int * int)
    = "dbll_adabas_get"
  (* sqlda_get s k = (sqltype, collength, colfrac).
   * where 'sqltype' is an encoding of several properties of a host
   * variable. see below.
   *)

external sqlda_get_int : sqlda -> int -> int
    = "dbll_adabas_get_int"
  (* gets integers if they are not too large *)

external sqlda_get_string : sqlda -> int -> string
    = "dbll_adabas_get_string"
  (* gets string *)

external sqlda_set_null : sqlda -> int -> unit
    = "dbll_adabas_set_null"

external sqlda_set_int : sqlda -> int -> int -> unit
    = "dbll_adabas_set_int"

external sqlda_set_string : sqlda -> int -> string -> unit
    = "dbll_adabas_set_string"

external sqlda_varinit : sqlda -> int -> unit
    = "dbll_adabas_varinit"


(* sqltype interpretation: 
 *
 * sqltype encoding:
 * bits 0-1: value of colio
 * bit  2:   value of colmode
 * bits 3-7: value of coltype
 * bits 8-9: abbreviated value of hostindicator:
 *   00: value is OK
 *   01: value is NULL
 *   10: value is not defined upon an error
 *   11: value is shortened
 *)

let sqltype_is_input k     = (k land 3) <> 1;;
let sqltype_is_output k    = (k land 3) <> 0;;
let sqltype_indicator k    = (k lsr 8) land 3;;
let sqltype_null_allowed k = ((k lsr 2) land 1) = 1;;
let sqltype_coltype k      = (k lsr 3) land 0x1f;;



(**********************************************************************)
(*                                                                    *)
(* date and time                                                      *)
(*                                                                    *)
(**********************************************************************)

exception Format_error;;


(* Note that "split" is a function not very specific to this particular
 * driver. Perhaps it is better to put it somewhere else
 *)

let split format str =
  (* format: 
   * special characters are '#' and '*', all other characters must match
   * literally. '#' matches digits; '*' matches the empty string, but
   * forces the previously matched number to be included in the result list
   * e.g. split "a#*###*abc" "a1234abc" yields [1; 234].
   * Raises Format_error if 'str' does not match 'format'.
   *)
  let l_f = String.length format in
  let l_s = String.length str in
  let rec spl i k v =
    if i >= l_f then begin
      if k >= l_s then [] else raise Format_error
    end
    else
      if i < l_f then begin
	let f = format.[i] in
	match f with
	  '#' -> 
	    if k >= l_s then raise Format_error;
	    let c = str.[k] in
	    begin match c with
	      '0'..'9' -> spl (i+1) (k+1) (v * 10 + Char.code(c) - 48)
	    | _        -> raise Format_error
	    end
	| '*' ->
	    v :: spl (i+1) k 0
	| _ ->
	    if k >= l_s then raise Format_error;
	    let c = str.[k] in
	    if c = f then
	      spl (i+1) (k+1) 0
	    else
	      raise Format_error
      end
      else raise Format_error
  in
  spl 0 0 0
;;


let rec first_of l str =
  match l with
    [] -> failwith "Dbll_adabas: Date/time format error"
  | f :: l' ->
      try
	f str
      with
	Format_error -> first_of l' str
;;
      

let date_to_string x =
  let (y,m,d) = yyyymmdd x in
  Printf.sprintf "%04d-%02d-%02d" y m d        (* ISO format *)
;;


let time_to_string x =
  let (h,m,s) = hhmmss x in
  Printf.sprintf "%02d.%02d.%02d" h m s        (* ISO format *)
;;


let timestamp_to_string (date,time,fraction) =
  let (y,m,d) = yyyymmdd date in
  let (h,min,s) = hhmmss time in
  let usecs = fraction / 1000 in
  Printf.sprintf "%04d-%02d-%02d-%02d.%02d.%02d.%06d" y m d h min s usecs
    (* ISO format *)
;;


let date_of_string s =           (* can scan every Adabas date format *)
  let (y,m,d) =
    first_of
      [ (fun s -> let [d; m; y] = split "##*.##*.####*" s in (y,m,d));
	(fun s -> let [y; m; d] = split "####*##*##*" s in (y,m,d));
	(fun s -> let [y; m; d] = split "####*-##*-##*" s in (y,m,d));
	(fun s -> let [m; d; y] = split "##*/##*/####*" s in (y,m,d)) ]
      s
  in
  make_date (y,m,d)
;;


let time_of_string s =           (* can scan every Adabas time format *)
  let (h,m,s) =
    first_of
      [ (fun s -> let [h; m; s] = split "##*.##*.##*" s in (h,m,s));
	(fun s -> let [h; m; s] = split "####*##*##*" s in (h,m,s));
	(fun s -> let [h; m; s] = split "##*:##*:##*" s in (h,m,s));
	(fun s -> let [h; m] = split "##*:##* AM" s in (h,m,0));
	(fun s -> let [h; m] = split "##*:##* PM" s in (h+12,m,0))
      ]
      s
  in
  make_time (h,m,s)
;;


let timestamp_of_string s =       (* can scan every Adabas timestamp format *)
  let (y,m,d,h,min,s,nanosec) =
    first_of
      [ (fun s -> let [y;m;d;h;min;s;usec] = 
	  split "####*-##*-##*-##*.##*.##*.######*" s in 
         (y,m,d,h,min,s,1000*usec));
	(fun s -> let [y;m;d;h;min;s;usec] = 
	  split "####*##*##*##*##*##*######*" s in 
         (y,m,d,h,min,s,1000*usec))
      ]	
      s
  in
  (make_date (y,m,d), make_time (h,min,s), nanosec)
;;
 

(**********************************************************************)
(*                                                                    *)
(* Interpretation of SQLDA fields                                     *)
(*                                                                    *)
(**********************************************************************)


let sig_of_coltype coltype = 
  match coltype with
     0 -> T_decimal
  |  1 -> T_float
  |  2 -> T_char
  |  3 -> T_byte
  |  4 -> T_date
  |  5 -> T_time
  |  7 -> T_float
  |  8 -> T_timestamp
  | 11 -> T_large_text
  | 12 -> T_large_bin
  | 15 -> T_smallint
  | 16 -> T_int
  | 17 -> T_varchar
  | 18 -> T_special "escapechar"
  | 19 -> T_large_text
  | 20 -> T_large_bin
  | 23 -> T_bool
  | _  -> failwith "Dbll_adabas: illegal column type"
;;


let sqlda_set_value s k v =
  match v with
    Decimal str    -> sqlda_set_string s k str
  | Float str      -> sqlda_set_string s k str
  | Int i          -> sqlda_set_int s k i
  | Bigint str     -> sqlda_set_string s k str
  | Date date      -> sqlda_set_string s k (date_to_string date)
  | Time time      -> sqlda_set_string s k (time_to_string time)
  | Timestamp (date, time, fraction) 
    -> sqlda_set_string s k (timestamp_to_string (date,time,fraction))
  | Byte str       -> sqlda_set_string s k str
  | Char str       -> sqlda_set_string s k str
  | Varchar str    -> sqlda_set_string s k str
  | Bool b         -> sqlda_set_int s k (if b then 1 else 0)
  | Large_text str -> sqlda_set_string s k str
  | Large_bin str  -> sqlda_set_string s k str
  | Unknown        -> failwith "Dbll_adabas: cannot set to unknown value"
  | Null           -> sqlda_set_null s k
;;


let sqlda_get_value s k =
  let (tp, _, _) = sqlda_get s k in
  let coltype = sqltype_coltype tp in

  match sqltype_indicator tp with
    1 -> Null
  | 2 -> Unknown
  | 3 -> Unknown
  | 0 -> 
      begin match coltype with
	 0 -> Decimal (sqlda_get_string s k)
      |  1 -> Float (sqlda_get_string s k)
      |  2 -> Char (sqlda_get_string s k)
      |  3 -> Byte (sqlda_get_string s k)
      |  4 -> Date (date_of_string (sqlda_get_string s k))
      |  5 -> Time (time_of_string (sqlda_get_string s k))
      |  7 -> Float (sqlda_get_string s k)
      |  8 -> Timestamp (timestamp_of_string (sqlda_get_string s k))
      | 11 -> Large_text (sqlda_get_string s k)
      | 12 -> Large_bin (sqlda_get_string s k)
      | 15 -> Int (sqlda_get_int s k)
      | 16 -> 
	  ( try 
	    Int (sqlda_get_int s k)
        with
	    Not_found -> 
	      Bigint (sqlda_get_string s k)
		)
      | 17 -> Varchar (sqlda_get_string s k)
      | 18 -> Unknown
      | 19 -> Large_text (sqlda_get_string s k)
      | 20 -> Large_bin  (sqlda_get_string s k)
      | 23 -> Bool ((sqlda_get_int s k) = 1)
      | _  -> failwith "Dbll_adabas: illegal column type"
      end
;;

let sqlda_count s p =
  let n = sqlda_length s in
  let rec count k =
    if k < n then begin
      let (tp, _, _) = sqlda_get s k in
      (if p tp then 1 else 0) + count (k+1)
    end
    else 0
  in
  count 0
;;


let run f =
  (* runs an SQL primitive and detects error conditions *)

  let result = f() in
  let code = sqlca_code() in
  match code with
    0   -> result   (* successful *)
  | 100 -> raise No_rows_found   (* row not found: this is not an error *)
  | (500 | -50 | -51 | -913)  -> raise Timeout_request
  | (700 | -70) -> raise Timeout_session
  | (600 | -60 | -911) -> raise Timeout_lock
  | (650 | 750) -> 
      raise (Transaction_error(code, sqlca_message()))
  | n when n > 0 ->
      raise (Error(code, sqlca_message()))
  | _ ->
      raise (Fatal_error(code, sqlca_message()))
;;


let fill sqlda a p =
  let k = ref 0 in
  for i = 0 to sqlda_length sqlda - 1 do
    let (tp, _, _) = sqlda_get sqlda !k in
    if p tp then begin
      let dbt = sig_of_coltype (sqltype_coltype tp) in
      a.( !k ) <- dbt;
      incr k
    end
  done
;;


(**********************************************************************)
(*                                                                    *)
(* Connection and Statement management                                *)
(*                                                                    *)
(**********************************************************************)

let max_connections = 4;;
    (* maximum number of simultaneous connections *)

let max_statements = 4;;
    (* maximum number of prepared statements *)

let connections =
  Array.make max_connections false
;;

type statement =
    { mutable serial_nr : int;
      mutable has_cursor : bool;
      mutable time_of_last_access : int
    } 
;;

let new_stm() =
  { serial_nr = -1;
    has_cursor = false;
    time_of_last_access = 0
  };;


let init_stm n =
  let a = Array.create n (new_stm()) in
  for i=0 to n-1 do
    a.(i) <- new_stm()
  done;
  a
;;



let next_connection() =
  let rec find n =
    if n = Array.length connections then
      raise Not_found
    else
      if connections.(n) then
	find (n+1)
      else
	n
  in
  find 0
;;


(**********************************************************************)
(*                                                                    *)
(* The driver object                                                  *)
(*                                                                    *)
(**********************************************************************)

class t =
  object (self) 

    inherit Dbll.t
      
    val mutable login = (None : Db_login.t option)
	(* If None, connection is closed. Otherwise, contains login info *)

    val mutable stm = init_stm max_statements
	(* for every statement number some info *)

    val mutable time = 0
	(* incremented on each "execute" *)

    val mutable cnr = -1
	(* connection number, from 0 to max_connections-1. -1 when
	 * not connected
	 *)

    val mutable serial_nr = 0
	(* next serial number *)

	(* TODO: logging, isolation level, timeouts *)

    method open_conn the_login =
      if login <> None then
	failwith "dbll_adabas: Connection already open";
      try 
	let n = next_connection() in

	(* Note that the "login" object may interactively ask the user
	 * about these values, so we have an *order*.
	 *)
	let host      = the_login # host() in
	let db        = the_login # database() in
	let user_name = the_login # user() in
	let password  = the_login # password() in

	let emulation_str = 
	  try 
	    the_login # option "adabas" "emulation" 
	  with
	    Not_found -> "adabas"
	in

	let isolation_level_str =
	  try
	    the_login # option "adabas" "isolation_level"
	  with
	    Not_found -> "serializable"
	in

	let session_timeout_str = 
	  try
	    the_login # option "adabas" "timeout_for_session"
	  with
	    Not_found -> "0"
	in

	let emulation =
	  match emulation_str with
	    "adabas" -> 0
	  | "oracle" -> 1
	  | "db2" -> 2
	  | "ansi" -> 3
	  | _ -> 0
	in

	let isolation_level =
	  match isolation_level_str with
	    "read uncommitted" -> 0
	  | "read committed"   -> 1
	  | "repeatable read"  -> 2
	  | "serializable"     -> 3
	  | _                  -> 3
	in

	let session_timeout = 
	  try int_of_string session_timeout_str 
	  with _ -> 0
	in

	run (fun () -> sql_connect n user_name password db host emulation isolation_level session_timeout);
	
	(* TODO: set adabas mode, timeout values and so on, too *)

	login <- Some the_login;
	cnr <- n;
	connections.(n) <- true;
	time <- 0;
	stm <- init_stm max_statements
      with
	Not_found ->                     
          (* next_connection() may raise this *)
	  failwith "dbll_adabas: too many connections"


    method close_conn () =
      if login = None then
	failwith "dbll_adabas: not connected";

      self # rollback_release();

      login <- None;
      connections.(cnr) <- false;
      cnr <- -1


    method private rollback_release() = 
      run (fun () -> sql_transaction cnr 3);
      ()

    method commit() =
      if login = None then
	failwith "dbll_adabas: not connected";
      run (fun () -> sql_transaction cnr 0);
      self # new_transaction();
      ()

    method rollback() =
      if login = None then
	failwith "dbll_adabas: not connected";
      run (fun () -> sql_transaction cnr 1);
      self # new_transaction();
      ()

    method private new_transaction() =
      (* cleanup internal state *)
      stm <- init_stm max_statements

    method prepare stm_name stm_text =
      if login = None then
	failwith "dbll_adabas: not connected";
      if stm.(stm_name).has_cursor then
	failwith "dbll_adabas: bad statement number, there is a cursor allocated for it";
      let statement = new_stm () in
      let s = serial_nr in
      stm.(stm_name) <- statement;   (* this stm number is now free again *)
      run (fun () -> sql_prepare cnr stm_name stm_text);
      statement.serial_nr <- s;      (* allocate stm number *)
      statement.time_of_last_access <- time;
      serial_nr <- s + 1;
      s

    method describe stm_name =
      (* describe stm_name and return signature *)
      if login = None then
	failwith "dbll_adabas: not connected";
      if stm.(stm_name).serial_nr = -1 then
	failwith "dbll_adabas: bad statement number";
      let sqlda    = run (fun () -> sql_describe cnr stm_name) in
      let n_input  = sqlda_count sqlda sqltype_is_input in
      let n_output = sqlda_count sqlda sqltype_is_output in
      let a_input  = Array.create n_input  T_int in
      let a_output = Array.create n_output T_int in

      fill sqlda a_input  sqltype_is_input;
      fill sqlda a_output sqltype_is_output;

      a_input, a_output

	
    method private setup_sqlda sqlda hostvar_input =
      (* copy input values from 'hostvar_input' into 'sqlda' structure: *)

      let k = ref 0 in
      for i = 0 to sqlda_length sqlda - 1 do
	let (tp, _, _) = sqlda_get sqlda !k in
	if sqltype_is_input tp then begin
	  sqlda_set_value sqlda i (hostvar_input.( !k ));
	  incr k
	end 
	else
	  sqlda_varinit sqlda i
      done


    method private read_sqlda sqlda =
      let n_output = sqlda_count sqlda sqltype_is_output in
      let a_output = Array.create n_output Unknown in

      let k = ref 0 in
      for i = 0 to sqlda_length sqlda - 1 do
	let (tp, _, _) = sqlda_get sqlda !k in
	if sqltype_is_output tp then begin
	  a_output.( !k ) <- sqlda_get_value sqlda i;
	  incr k
	end
      done;

      a_output


    method execute stm_name hostvar_input =
      if login = None then
	failwith "dbll_adabas: not connected";
      if stm.(stm_name).serial_nr = -1 then
	failwith "dbll_adabas: bad statement number";
      time <- time + 1;
      stm.(stm_name).time_of_last_access <- time;
      let sqlda = run (fun () -> sql_describe cnr stm_name) in
      self # setup_sqlda sqlda hostvar_input;
      run (fun () -> sql_execute cnr stm_name sqlda);
      self # read_sqlda sqlda


    method query stm_name hostvar_input =
      if login = None then
	failwith "dbll_adabas: not connected";
      if stm.(stm_name).serial_nr = -1 then
	failwith "dbll_adabas: bad statement number";
      time <- time + 1;
      stm.(stm_name).time_of_last_access <- time;
      let sqlda = run (fun () -> sql_describe cnr stm_name) in
      self # setup_sqlda sqlda hostvar_input;
      run (fun () -> sql_declare_cursor cnr stm_name);
      run (fun () -> sql_open_cursor cnr stm_name sqlda);
      stm.(stm_name).has_cursor <- true;
      self # read_sqlda sqlda


    method describe_query stm_name =
      if login = None then
	failwith "dbll_adabas: not connected";
      if stm.(stm_name).serial_nr = -1 then
	failwith "dbll_adabas: bad statement number";
      if not stm.(stm_name).has_cursor then
	failwith "dbll_adabas: no cursor";
      let sqlda    = run (fun () -> sql_describe_fetch_cursor cnr stm_name) in
      let n_input  = sqlda_count sqlda sqltype_is_input in
      let n_output = sqlda_count sqlda sqltype_is_output in

      if n_input > 0 then
	failwith "dbll_adabas: assertion n_input = 0 violated";

      let a_output = Array.create n_output T_int in

      fill sqlda a_output sqltype_is_output;

      a_output


    method fetch stm_name =
      if login = None then
	failwith "dbll_adabas: not connected";
      if stm.(stm_name).serial_nr = -1 then
	failwith "dbll_adabas: bad statement number";
      if not stm.(stm_name).has_cursor then
	failwith "dbll_adabas: no cursor";
      let sqlda = run (fun () -> sql_describe_fetch_cursor cnr stm_name) in
      let rec next () =
	self # setup_sqlda sqlda [| |];
	try
	  run (fun () -> sql_fetch_cursor cnr stm_name sqlda);
	  let result = self # read_sqlda sqlda in
	  [< ' result; next() >]
	with
	  No_rows_found -> 
	    [< >]
      in
      [< next() >]


    method cursor_name stm_name =
      if login = None then
	failwith "dbll_adabas: not connected";
      if stm.(stm_name).serial_nr = -1 then
	failwith "dbll_adabas: bad statement number";
      if not stm.(stm_name).has_cursor then
	failwith "dbll_adabas: no cursor";
      "dbll_c" ^ (string_of_int cnr) ^ (string_of_int stm_name)


    method release stm_name =
      if login = None then
	failwith "dbll_adabas: not connected";
      if stm.(stm_name).serial_nr = -1 then
	failwith "dbll_adabas: bad statement number";
      if not stm.(stm_name).has_cursor then
	failwith "dbll_adabas: no cursor";
      run (fun () -> sql_close_cursor cnr stm_name);
      stm.(stm_name).has_cursor <- false


    method state_is_open () =
      login <> None


    method state_maximum_number_of_statements () =
      max_statements


    method state_serial_number stm_name =
      if login = None then
	failwith "dbll_adabas: not connected";
      stm.(stm_name).serial_nr


    method state_has_cursor stm_name =
      if login = None then
	failwith "dbll_adabas: not connected";
      stm.(stm_name).has_cursor


    method state_next_stm_number () =
      (* get the statement number with lowest serial nr that does not have
       * a cursor
       *)
      let rec find best_stm_nr best_serial_nr k =
	if k >= max_statements then begin
	  if best_stm_nr = -1 then
	    raise Not_found
	  else
	    best_stm_nr
	end
	else
	  if not stm.(k).has_cursor then begin
	    if best_stm_nr = -1 then
	      find k stm.(k).serial_nr (k+1)
	    else
	      if stm.(k).serial_nr < best_serial_nr then
		find k stm.(k).serial_nr (k+1)
	      else
		find best_stm_nr best_serial_nr (k+1)
	  end
	  else
	    find best_stm_nr best_serial_nr (k+1)
      in
      find (-1) 0 0

  end

(* ======================================================================
 * History:
 * 
 * $Log: dbll_adabas.ml,v $
 * Revision 1.2  1999/01/28 02:06:36  gerd
 * Adabas driver:
 * 	- added option "adabas.timeout_for_session"
 * 	- added option "adabas.isolation_level"
 * 	- tried to add "adabas.emulation" to set the SQLMODE. There is a
 * 	  serious problem since the Adabas documentation describes the
 * 	  SQLMODE parameter of CONNECT which seems not to work
 *
 * Revision 1.1  1999/01/26 02:10:57  gerd
 * first release
 *
 * 
 *)
