(* $Id: db.ml,v 1.2 1999/01/27 20:08:17 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

open Db_types


type statement =
    { mutable stm_nr : int;                  (* Dbll statement number *)
      mutable stm_serial : int;              (* Dbll serial number *)
              stm_text : string;             (* the statement itself *)
              stm_in_sig : signature array;  (* input signature *)
              stm_out_sig : signature array; (* output signature *)
    } 
;;


type result = 
    {         cursor_stm : statement;
      mutable cursor_is_open : bool;
              cursor_out_sig : signature array
    } 
;;


class t (the_driver:Dbll.t) =
  object (self)
    val driver = the_driver

    val mutable p_open = false

    method is_open () = p_open

    method connect login =
      driver # open_conn login;
      p_open <- true

    method close () =
      driver # commit();
      driver # close_conn();
      p_open <- false

    method abort_connection () =
      driver # close_conn();
      p_open <- false

    method private internal_abort () =
      try
	driver # close_conn();
	p_open <- false
      with
	any -> ()

    method commit () =
      try
	driver # commit();
      with
       	Fatal_error x ->
	  self # internal_abort();
	  raise (Fatal_error x)

    method rollback () =
      try
	driver # rollback()
      with
      	Fatal_error x ->
	  self # internal_abort();
	  raise(Fatal_error x)


    method private re_prepare s =
      let c_serial = driver # state_serial_number s.stm_nr in
      if c_serial <> s.stm_serial then begin
	let nr = 
	  try 
	    driver # state_next_stm_number()
	  with
	    Not_found ->
	      failwith "Db: cannot prepare statement again because of too many open cursors"
	in
	let serial = driver # prepare nr s.stm_text in
	s.stm_nr <- nr;
	s.stm_serial <- serial
      end


    method prepare_string (stm:string) =
      try
	let nr = 
	  try 
	    driver # state_next_stm_number()
	  with
	    Not_found ->
	      failwith "Db: cannot prepare statement because of too many open cursors"
	in
	let serial = driver # prepare nr stm in
	let in_sig, out_sig = driver # describe nr in
	let s =
	  { stm_nr = nr;
	    stm_serial = serial;
	    stm_text = stm;
	    stm_in_sig = in_sig;
	    stm_out_sig = out_sig
	  }  in

	s
	  
      with
      	Fatal_error x as ex ->
	  self # internal_abort();
	  raise ex


    method description (stm:statement) =
      stm.stm_in_sig, stm.stm_out_sig

    method execute (stm:statement) (vals:value array) : value array =
      try
	self # check_sig stm.stm_in_sig vals;
	self # re_prepare stm;
	driver # execute stm.stm_nr vals
      with
      	Fatal_error x as ex ->
	  self # internal_abort();
	  raise ex

    method select (stm:statement) (vals:value array) : result =
      try
	self # check_sig stm.stm_in_sig vals;
	self # re_prepare stm;
	let r = driver # query stm.stm_nr vals in
	try
	  if r <> [| |] then
	    failwith "Db: strange 'query' result";
	  let d = driver # describe_query stm.stm_nr  in
	  let c = 
	    { cursor_stm = stm;
	      cursor_is_open = false;
	      cursor_out_sig = d
	    } in
	  c
	with
	  any ->
	    driver # release stm.stm_nr;
	    raise any
      with
      	Fatal_error x as ex ->
	  self # internal_abort();
	  raise ex


    method open_result (r:result) : value array Stream.t =
      try
	if r.cursor_is_open then
	  failwith "Db: cannot open the cursor again";
	let s = driver # fetch (r.cursor_stm.stm_nr) in
	r.cursor_is_open <- true;
	s
      with
      	Fatal_error x as ex ->
	  self # internal_abort();
	  raise ex


    method close_result (r:result) : unit =
      try
	if not r.cursor_is_open then
	  failwith "Db: cursor not open";
	driver # release (r.cursor_stm.stm_nr)
	(* r.cursor_is_open <- false *)
      with
      	Fatal_error x as ex ->
	  self # internal_abort();
	  raise ex


    method private check_sig sign vals =
      let error() =
	failwith "Db: values do not match signature"
      in
      if Array.length sign <> Array.length vals then
	error();
      for i = 0 to Array.length vals - 1 do
	let t = sign.(i) in
	let x = vals.(i) in
	match x with
	  Decimal _    ->  if t <> T_decimal & t <> T_int & t <> T_smallint 
	                   then error()
	| Float _      ->  if t <> T_float   & t <> T_int & t <> T_smallint 
	                   then error()
	| Int _        ->  if t <> T_int & t <> T_smallint then error()
	| Bigint _     ->  if t <> T_int & t <> T_smallint then error()
	| Date _       ->  if t <> T_date then error()
	| Time _       ->  if t <> T_time then error()
	| Timestamp _  ->  if t <> T_timestamp then error()
	| Byte _       ->  if t <> T_byte then error()
	| Char _       ->  if t <> T_char then error()
	| Varchar _    ->  if t <> T_varchar then error()
	| Bool _       ->  if t <> T_bool then error()
	| Large_text _ ->  if t <> T_large_text then error()
	| Large_bin _  ->  if t <> T_large_bin then error()
	| Unknown      ->  error()
	| Null         ->  ()          (* check on NULL not implemented *)
      done

    method do_string (str:string) (vals:value array) : unit =
      let stm = self # prepare_string str in
      self # execute stm vals;
      ()

    method do_file filename vals =
      let str = self # read_file filename in
      self # do_string str vals

    method query_string (str:string) (vals:value array) : value array list =
      let stm = self # prepare_string str in
      let result = self # select stm vals in
      let stream = self # open_result result in
      try
	let l = self # list_of_stream stream in
	self # close_result result;
	l
      with
	any ->
	  self # close_result result;
	  raise any
      
    method query_file filename vals =
      let str = self # read_file filename in
      self # query_string str vals

    method private list_of_stream stream =
      let rec read () =
	try
	  Stream.next stream :: read()
	with
	  Stream.Failure -> []
      in
      read()

    method private read_file filename =
      let ch = open_in filename in
      let sl = ref [] in
      try
	while true do
	  (* TODO: remove --comments *)
	  sl := !sl @ [ input_line ch ]
	done;
	failwith "Ocaml does not compile correctly"
      with
	End_of_file ->
	  close_in ch;
	  String.concat "\n" !sl
      |	any ->
	  close_in ch;
	  raise any

  end
;;

(* ======================================================================
 * History:
 * 
 * $Log: db.ml,v $
 * Revision 1.2  1999/01/27 20:08:17  gerd
 * Db.mli:
 * 	lquery_string is now query_string; there is not anymore
 *  	an immediately executing method that returns a stream, because
 * 	you can't release the cursor after the result has been read.
 *
 * 	added query_file and do_file that execute statements stored
 * 	in files
 * Db.ml:
 * 	implemented {query,do}_{string,file}. The implementation of
 * 	Db.ml is now complete.
 *
 * Revision 1.1  1999/01/26 02:10:53  gerd
 * first release
 *
 * 
 *)
