(* $Id: db.mli,v 1.2 1999/01/27 20:08:18 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(**********************************************************************)
(*                                                                    *)
(* GENERAL DBMS ACCESS INTERFACE                                      *)
(*                                                                    *)
(**********************************************************************)

open Db_types

type statement;;
type result;;


(**************************** THE OBJECT *****************************)

class t :
    (Dbll.t) ->
    object
	

      (********************* CONNECTION *********************)

      (* This object represents a usually open database connection
       * to a certain DBMS. More specific, the connection is either
       * open or stale (closed, broken or unusuable because of another
       * reason)
       *)

      method is_open : unit -> bool

      method connect : Db_login.t -> unit

      method close : unit -> unit
	  (* does a 'commit' and then closes the connection *)

      method abort_connection : unit -> unit
	  (* does a 'rollback' and then closes the connection *)


      (********************* TRANSACTIONS *************************)

      (* Note that:
       * - it must already be indicated on creation of this object if logging
       *   should be done or not. This cannot be changed later.
       * - the isolation level must be given on creation, too
       *)

      method commit : unit -> unit

      method rollback : unit -> unit


      (********************** EXECUTION ***************************)

      (* prepare_string: prepares an SQL statement that is given as
       * string. Host variables are indicated by '?' characters
       * (on legal positions, that depends on the DBMS). Returns
       * a description which variables do input and output, and
       * what type the variables have (the signature list). The
       * statement denotes the prepared statement for later
       * execution.
       *)

      method prepare_string : string -> statement

      method description : statement -> (signature array * signature array)

      (* execute: does an already prepared statement. if host variables
       * are used, the concrete input values must be given, too.
       *)

      method execute : statement -> value array -> value array

      method select : statement -> value array -> result

      (* open_result: returns the stream of the result table. This method
       * can only be invoked on already executed statements. It implicitly
       * opens a cursor that remains open until (i) the stream has been
       * completely read (i.e. until the end of the stream could be detected)
       * or (ii) close_result is invoked or (iii) the connection ends.
       *
       * Note: If you invoke 'open_result' again after a part of the result has
       * been read the behaviour is undefined.
       *)

      method open_result : result -> value array Stream.t

      method close_result : result -> unit


      (********************* MORE COMFORT ****************************)

      method do_string : string -> value array -> unit

	  (* do_string cmd vals: 
	   * prepares and executes the statement 'cmd' with input host
	   * variables 'vals'; output values are discarded.
	   *)

      method do_file   : string -> value array -> unit

	  (* do_file filename vals:
	   * same as 'do_string', but the statement is read from the file
	   * denoted by 'filename'
	   *)

      method query_string : string -> value array -> value array list

	  (* query_string cmd vals:
	   * prepares and executes as "select" statement. The result table
	   * is returned as list.
	   *)

      method query_file   : string -> value array -> value array list

	  (* query_file filename vals:
	   * same as 'query_string', but the statement is read from the file
	   * denoted by 'filename'
	   *)


      (********************* IDEAS FOR FUTURE *************************)

      (* prepare_sql : sql -> (signature array * signature array * statement)
       * 
       * where 'sql' is a term describing a valid expression in SQL 92.
       * Useful to better manipulate SQL expressions. There should also be
       * a parser to get an 'sql' expression from a string. This is a
       * way to get stricter SQL92 compatibility.
       *)

    end
;;

(* ======================================================================
 * History:
 * 
 * $Log: db.mli,v $
 * Revision 1.2  1999/01/27 20:08:18  gerd
 * Db.mli:
 * 	lquery_string is now query_string; there is not anymore
 *  	an immediately executing method that returns a stream, because
 * 	you can't release the cursor after the result has been read.
 *
 * 	added query_file and do_file that execute statements stored
 * 	in files
 * Db.ml:
 * 	implemented {query,do}_{string,file}. The implementation of
 * 	Db.ml is now complete.
 *
 * Revision 1.1  1999/01/26 02:10:54  gerd
 * first release
 *
 * 
 *)
