(* $Id: db_login.ml,v 1.2 1999/01/27 19:30:04 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(**********************************************************************)
(* LOGIN OBJECTS                                                      *)
(*                                                                    *)
(* Contains necessary login data for DBMS access.                     *)
(*                                                                    *)
(**********************************************************************)

open Genlex

type login_term =
    { 
      (* vital data *)

      mutable user     : login_string;
      mutable password : login_string;
      mutable host     : login_string;
      mutable database : login_string;

      mutable opt      : (string * login_string) list;

      (* opt: first component is a string of the form "<dbms>.<property>"
       * where <dbms> is the name of the DBMS driver or "*" to address all
       * drivers, and <property> the name of the property.
       *
       * Standard properties recognized by many drivers:
       *
       * "trace": the value is the name of a trace file
       * "max_statements": the maximum number of statement that can be
       *    prepared simultaneously (if this number if configurable)
       * "min_statements": the minimum number of statements that must be
       *    preparable to run this application. The driver generates an
       *    error if it cannot fulfill this value.
       * "timeout_for_requests": A timeout value in seconds for requests that
       *    wait too long for locks owned by others
       * "timeout_for_locks": A timeout value in seconds for locks owned
       *    by this session
       * "timeout_for_session": A global timeout value for the duration of
       *    this session (in seconds, again)
       *)
  } 

and login_string =
    Constant of string
  | Current_unix_user
  | Interactive of string                 (* The string is the prompt *)
  | Invisible_interactive of string       (* similar, but for passwords *)
  | Off                                   (* this option is turned off *)
;;


type match_result =
    Matches_exactly
  | Matches_generally
  | Matches_specially
  | Does_not_match
;;


let no_login() =
  { user = Off;
    password = Off;
    host = Off;
    database = Off;
    opt = []
  } 
;;


(* parse dumped files: *)

let rec parse_file = parser
    [< s = parse_setting; rest = parse_file >] ->
      if rest.user <> Off then
	s.user <- rest.user;
      if rest.password <> Off then
	s.password <- rest.password;
      if rest.host <> Off then
	s.host <- rest.host;
      if rest.database <> Off then
	s.database <- rest.database;
      s.opt <- s.opt @ rest.opt;
      s
  | [< >] ->
      no_login()

and parse_setting = parser
    [< 'Kwd "user";     'Kwd "="; value = parse_value >] ->
      let x = no_login() in
      x.user <- value;
      x
  | [< 'Kwd "password"; 'Kwd "=";  value = parse_value >] ->
      let x = no_login() in
      x.password <- value;
      x
  | [< 'Kwd "host";     'Kwd "=";  value = parse_value >] ->
      let x = no_login() in
      x.host <- value;
      x
  | [< 'Kwd "database"; 'Kwd "=";  value = parse_value >] ->
      let x = no_login() in
      x.database <- value;
      x
  | [< 'Kwd "option";   'String key;  'Kwd "=";  value = parse_value >] ->
      let x = no_login() in
      x.opt <- [ key, value ];
      x

and parse_value = parser
    [< 'Kwd "interactive"; 'String s>] -> Interactive s
  | [< 'Kwd "invisible"; 'Kwd "interactive"; 'String s >] -> 
      Invisible_interactive s
  | [< 'Kwd "unixuser" >] -> Current_unix_user
  | [< 'Kwd "off" >] -> Off
  | [< 'String s >] -> Constant s
  | [< 'Int n >] -> Constant (string_of_int n)
;;



class t =
  object (self) 

    val mutable data = no_login()

    (***** query access data *****)

    method user () =
      (* returns the string or Not_found. Interactive dialogs are done now. *)
      let s = self # get data.user in
      data.user <- Constant s;
      s

    method password () =
      (* returns the string or Not_found. Interactive dialogs are done now. *)
      let s = self # get data.password in
      data.password <- Constant s;
      s

    method host () =
      (* returns the string or Not_found. Interactive dialogs are done now. *)
      let s = self # get data.host in
      data.host <- Constant s;
      s

    method database () =
      (* returns the string or Not_found. Interactive dialogs are done now. *)
      let s = self # get data.database in
      data.database <- Constant s;
      s

    method option dbms key =
      let pattern,s = self # get_option dbms key in
      self # set_option pattern (Constant s);
      s

    method private get value =
      match value with
	Constant s -> s
      |	Current_unix_user -> 
	  Unix.getlogin()
      |	Interactive prompt ->
	  print_string prompt;
	  read_line()
      |	Invisible_interactive prompt ->
	  self # read_invisible_string prompt
      |	Off ->
	  raise Not_found
	  
    method private get_option dbms key =
      let k = dbms ^ "." ^ key in

      let rec find_special l =
	match l with
	  [] -> raise Not_found
	| (key_p, value_p) :: l' ->
	    if self # match_option k key_p = Matches_exactly then
	      key_p, self # get value_p
	    else
	      find_special l'
      in

      let rec find_general l =
	match l with
	  [] -> raise Not_found
	| (key_p, value_p) :: l' ->
	    if self # match_option k key_p = Matches_specially then
	      key_p, self # get value_p
	    else
	      find_general l'
      in

      try
	find_special data.opt
      with
	Not_found ->
	  find_general data.opt


    method private read_invisible_string prompt =
      print_string prompt;
      flush stdout;
      let old_terminal_mode = Unix.tcgetattr Unix.stdin in
      let terminal_mode = Unix.tcgetattr Unix.stdin in

      terminal_mode.Unix.c_icanon <- false; (* non-canonical mode *)
      terminal_mode.Unix.c_echo <- false;   (* no echo *)
      terminal_mode.Unix.c_vmin <- 1;       (* read at least 1 char at a time *)
      terminal_mode.Unix.c_vtime <- 0;      (* wait endless for characters *)

      try
	Unix.tcsetattr Unix.stdin Unix.TCSAFLUSH terminal_mode;
	
	let input = ref "" in
	let end_of_input = ref false in

	while not !end_of_input do
	  let c =
	    try
	      input_char stdin 
	    with
	      End_of_file ->
		'\n'
	  in
	  if c = '\n' or 
	     c = terminal_mode.Unix.c_veol or 
	     c = terminal_mode.Unix.c_veof then begin
	       print_newline();
	       flush stdout;
	       end_of_input := true
	     end
	  else
	    if c = terminal_mode.Unix.c_verase then begin
	      if !input <> "" then begin
		print_string "\008\032\008";
		flush stdout;
		input := String.sub !input 0 (String.length !input - 1)
	      end
	    end
	    else begin
	      print_char '*';
	      flush stdout;
	      input := !input ^ (String.make 1 c)
	    end
	done;
	Unix.tcsetattr Unix.stdin Unix.TCSAFLUSH old_terminal_mode;
	!input
      with
	any ->
	  Unix.tcsetattr Unix.stdin Unix.TCSAFLUSH old_terminal_mode;
	  raise any


    (***** set access data *****)


    method set_user user =
      data.user <- user

    method set_password passwd =
      data.password <- passwd

    method set_host host =
      data.host <- host

    method set_database db =
      data.database <- db


    method set_option key value =
      let rec add l =
	match l with
	  [] -> [ key, value ]
	| (key_c, value_c) :: l' -> 
	    if key = key_c then 
	      (key, value) :: l' 
	    else 
	      (key_c, value_c) :: add l'
      in
      data.opt <- add data.opt


    method private match_option pattern1 pattern2 =
      (* result:
       * Matches_exactly   if pattern1 = pattern2
       * Matches_generally if pattern1 is "*.key" and pattern2 is "x.key"
       * Matches_specially if pattern1 is "x.key" and pattern2 is "*.key"
       * Does_not_match    else
       *)
      let prefix s =
	try
	  String.sub s 0 (String.index s '.')
	with
	  Not_found -> ""
      in

      let suffix s =
	try
	  let n = String.index s '.' in
	  String.sub s (n+1) (String.length s - n - 1)
	with
	  Not_found -> s
      in


      if pattern1 = pattern2 then
	Matches_exactly
      else
	let prefix1 = prefix pattern1 in
	let prefix2 = prefix pattern2 in
	let suffix1 = suffix pattern1 in
	let suffix2 = suffix pattern2 in

	if suffix1 = suffix2 then begin
	  if prefix1 = "*" then 
	    Matches_generally
	  else if prefix2 = "*" then
	    Matches_specially
	  else
	    Does_not_match
	end
	else Does_not_match


    (***** read settings from file *****)


    method init_from_file filename =
      let ch = open_in filename in
      try
	let file_stream = Stream.of_channel ch in
	let lexer = Genlex.make_lexer [ "user";
					"password";
					"host";
					"database";
					"option";
					"interactive";
					"invisible";
					"unixuser";
					"off";
					"=" ]
	    file_stream
	in
	data <- parse_file lexer;
	Stream.empty lexer;
	close_in ch
      with
	Stream.Error s ->
	  close_in ch;
	  failwith ("error parsing file " ^ filename)
      |	Stream.Failure ->
	  close_in ch;
	  failwith ("error parsing file " ^ filename)
      |	Parsing.Parse_error ->
	  close_in ch;
	  failwith ("error parsing file " ^ filename)
      |	any ->
	  close_in ch;
	  raise any


    method init_from_string config =
      try
	let file_stream = Stream.of_string config in
	let lexer = Genlex.make_lexer [ "user";
					"password";
					"host";
					"database";
					"option";
					"interactive";
					"invisible";
					"unixuser";
					"off";
					"=" ]
	    file_stream
	in
	data <- parse_file lexer;
	Stream.empty lexer;
      with
	Stream.Error s ->
	  failwith ("error parsing configuration")
      |	Stream.Failure ->
	  failwith ("error parsing configuration")
      |	Parsing.Parse_error ->
	  failwith ("error parsing configuration")
  end

(* ======================================================================
 * History:
 * 
 * $Log: db_login.ml,v $
 * Revision 1.2  1999/01/27 19:30:04  gerd
 * - added db_login.mli
 * - added the method init_from_string in Db.login.t
 *
 * Revision 1.1  1999/01/26 02:10:54  gerd
 * first release
 *
 * 
 *)
