(* $Id: db_login.mli,v 1.1 1999/01/27 19:30:05 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(**********************************************************************)
(* LOGIN OBJECTS                                                      *)
(*                                                                    *)
(* Contains necessary login data for DBMS access.                     *)
(*                                                                    *)
(**********************************************************************)

type login_string =
    Constant of string
  | Current_unix_user
  | Interactive of string                 (* The string is the prompt *)
  | Invisible_interactive of string       (* similar, but for passwords *)
  | Off                                   (* this option is turned off *)


class t :
  object

    method user : unit -> string
	(* returns the current user or Not_found. The invocation of this
	 * method may effect an interactive dialog that prompts for the
	 * user.
	 *)

    method password : unit -> string
	(* returns the password or Not_found. The invocation of this
	 * method may effect an interactive dialog that prompts for the
	 * user.
	 *)

    method host : unit -> string
	(* returns the host name where the DBMS server runs or Not_found. 
	 * The invocation of this
	 * method may effect an interactive dialog that prompts for the
	 * user.
	 *)

    method database : unit -> string
	(* returns the database name or Not_found. The invocation of this
	 * method may effect an interactive dialog that prompts for the
	 * user.
	 *)

    method option : string -> string -> string
	(* option dbms_name option_name:
	 * Returns the value of the option 'option_name' for the DBMS
	 * 'dbms_name' or Not_found.
	 * If there is a special entry "<dbms_name>.<option_name>" this
	 * entry is returned. If there is a generic entry "*.<option_name>"
	 * this entry is returned. Otherwise, Not_found is raised.
	 *)

    method set_user : login_string -> unit
	(* Sets the user. Overrides previous setting if there is one. *)
	
    method set_password : login_string -> unit
	(* Sets the password. Overrides previous setting if there is one. *)

    method set_host : login_string -> unit
	(* Sets the host. Overrides previous setting if there is one. *)

    method set_database : login_string -> unit
	(* Sets the database. Overrides previous setting if there is one. *)

    method set_option : string -> login_string -> unit
	(* set_option option_spec value:
	 * Sets the option given by
	 *    'option_spec' = "<dbms_spec>.<option_name>"
	 * <dbms_spec> maybe "*" or the name of the DBMS to which this option
	 * applies.
	 * <option_name> is the name of the option.
	 * 'value' is an arbitrary string.
	 *)

    method init_from_file : string -> unit
	(* Sets user, password, host, database and all options as given
	 * in the configuration file whose name is passed as argument.
	 * This overrides all previous settings.
	 * Format of this file: see below 
	 *)

    method init_from_string : string -> unit
	(* As "init_from_file", but the configuration is passed directly
	 * as string.
	 *)

  end


(* Format of the configuration:
 *
 * The configuration is tokenized following the OCaml rules; that means
 * especially that white space may be anything of blanks, tabs or newlines
 * and that comments are indicated by (* ... *) and may be nested.
 *
 * A configuration is a sequence of settings. Possible settings:
 *
 * <setting> ::=
 *   user = <value>
 * | password = <value>
 * | host = <value>
 * | database = <value
 * | option <key> = <value>
 *
 * A <key> is just a string enclosed in "..." and should be of the format
 * "<dbms_spec>.<option_name>". The <dbms_spec> is the name of the DBMS
 * to which the option applies, or "*" if the option applies to any
 * DBMS that can interprete it.
 * 
 * A <value> may be any of the following:
 *
 * <value> ::=
 *   interactive "<prompt>"
 * | invisible interactive "<prompt>"
 * | unixuser
 * | off
 * | "<literal>"
 *
 * Here "<prompt>" and "<literal>" are string literals enclosed in double
 * quotes.
 * - interactive: forces that the prompt is printed to stdout and the 
 *   value can be entered interactively.
 * - invisible interactive: forces that the prompt is printed to stdout and 
 *   the value can be entered interactively, but is hidden
 * - unixuser: the current login name of the Unix user. Note that this can
 *   be different from the current UID/EUID and can be even empty.
 * - off: The option is turned off
 * - "<literal>": The option is set to this value.
 *
 * EXAMPLE:
 *
 * user = "somebody"
 * password = invisible interactive "Please enter your password: "
 * host = "localhost" database = "foo"
 * option "*.timeout_request" = 60
 * option "SlowDBMS.timeout_request" = 360
 * option "*.trace" = "/tmp/trace.log"
 *)


(* ======================================================================
 * History:
 * 
 * 
 *)
