(* $Id: db_types.mli,v 1.1 1999/01/26 02:10:54 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(**********************************************************************)
(*                                                                    *)
(* TYPES FOR THE GENERAL DBMS ACCESS INTERFACE                        *)
(*                                                                    *)
(**********************************************************************)


type signature =
    T_decimal             (* fixed-point number *)
  | T_float               (* floating point number *)
  | T_int                 (* 32 bit signed integer *)
  | T_smallint            (* 16 bit signed integer *)
  | T_date                (* date *)
  | T_time                (* time *)
  | T_timestamp           (* date + time + even more precise *)
  | T_byte                (* a binary string *)
  | T_char                (* a text of fixed length *)
  | T_varchar             (* a text of variable but limited length *)
  | T_bool                (* a boolean value *)
  | T_large_text          (* a text of arbitrary length (text blob) *)
  | T_large_bin           (* a binary string of arbitrary length (blob) *)
  | T_special of string   (* a DBMS specific type *)
;;


(* 'date' and 'time' are abstract to make it simpler to adapt it to
 * a universal date/time type if one writes such a module
 *)
type date;;
type time;;

type value =
    Decimal of string     (* decimal representation *)
  | Float of string       (* decimal representation *)
  | Int of int            (* only if number fits in 31 bits *)
  | Bigint of string
  | Date of date          (* year, month, day *)
  | Time of time          (* hours, minutes, seconds *)
  | Timestamp of (date * time * int)      (* and nanoseconds as third *)
  | Byte of string        (* a string of octets *)
  | Char of string        (* a string of characters of a character set *)
  | Varchar of string     (* a variable string of chars of a character set *)
  | Bool of bool          (* yes/no *)
  | Large_text of string  (* many *)
  | Large_bin of string   (* many *)
  | Unknown               (* not representable DBMS specific value *)
  | Null
;;


(* date/time access *)

val yyyymmdd : date -> (int * int * int);;
val hhmmss : time -> (int * int * int);;
val make_date : (int * int * int) -> date;;
val make_time : (int * int * int) -> time;;


(*************************** EXCEPTIONS ******************************)


exception No_rows_found;;


(* statement waits too long that somebody else releases a lock: *)
exception Timeout_request;;     

(* statement occupies a lock too long: 
 * If you catch this exception, do a 'rollback' (the driver should do that
 * for you, but this is not strictly specified)
 *)
exception Timeout_lock;;

(* session's idle time is too long: 
 * It is guaranteed that the connection has been closed.
 *)
exception Timeout_session;;

(* Error: the first component is the error number, then follows the error
 * text. An error indicates that the current statement could not be
 * executed (normal operating condition)
 *)

exception Error of (int * string)
;;

(* Transaction_error: same components. The error was so hard that the
 * current transaction has been rolled back.
 *)

exception Transaction_error of (int * string)
;;

(* Fatal_error: same components. A fatal error means that the connection
 * cannot be made or becomes unusable. A fatal error is also raised if there
 * are syntax errors in your SQL statements.
 *)

exception Fatal_error of (int * string)
;;

(* ======================================================================
 * History:
 * 
 * $Log: db_types.mli,v $
 * Revision 1.1  1999/01/26 02:10:54  gerd
 * first release
 *
 * 
 *)
