(* $Id: dbll.ml,v 1.1 1999/01/26 02:10:54 gerd Exp $
 * ----------------------------------------------------------------------
 *
 *)

(**********************************************************************)
(* LOW-LEVEL DATABASE LAYER                                           *)
(*                                                                    *)
(* This module descibes the abstract class Dbll.t that can be used    *)
(* as pattern for low-level drivers. Any driver must conform to       *)
(* Dbll.t and should therefore be simply derived from it.             *)
(*                                                                    *)
(**********************************************************************)

(* - connections: A driver must at least allow to open one connection.
 *   Better drivers allow to use several connections at the same time.
 *   A connection identifies a certain DBMS server, a certain database,
 *   the user that controls the connection, and the overall mode of the
 *   accesses.
 *   These parameters of the connection are passed to the driver by a
 *   Db_login.t object. The set of properties that such an object can
 *   express is extensible, which might be useful for a driver designer
 *   since it is an easy method to pass driver-specific additional
 *   connection parameters.
 *
 * - prepared statements: In order to execute a statement it must have
 *   been prepared earlier. The driver defines a maximum number of
 *   statements that can be prepared at the same time (for some drivers,
 *   this maximum can be set by the user). If you want to execute more
 *   statements a previously prepared statement must be given up. For
 *   simplicity, the driver enumerates all prepared statements from 0
 *   to N-1 (where N is the maximum), the statement number.
 *   All drivers should implement N >= 4.
 *   -- The driver automatically chooses names for statements. These
 *   names have the prefix "dbll_".
 * 
 *   Note: even if it is simple for the driver to allocate arbitrary numbers
 *   of statements, there should be a (perhaps configurable) limit; otherwise
 *   memory is wasted.
 *
 * - statements with cursor: For some sorts of statements you can define
 *   a cursor. The driver has a special interface that allows easy usage
 *   of cursors. If you use this interface, at most one cursor is possible
 *   per statement (but the driver assigns a name to the cursor, so you
 *   do not must do it yourself). The driver then opens a stream (of type
 *   Stream.t) that can be used for reading or writing. After all is done,
 *   (i.e. a part or the whole stream has been read), the cursor must be
 *   closed. 
 *   As streams complicate the control flow of the program, statement numbers
 *   with open cursors are protected against preparation. You have to
 *   explicitly close a cursor to reuse such a statment number. (Of course,
 *   opening another cursor for the same statement is forbidden, too.)
 *   -- The driver automatically chooses names for cursors managed by itself. 
 *   These names have the prefix "dbll_".
 *
 * - typechecks: After a statement has been prepared, it can be described.
 *   For every question mark it is decided whether it is an input or
 *   output parameter, and which type it has. The result of a description is
 *   a pair (signature list * signature list) where the first component
 *   refers to the input and the second component to the output parameters,
 *   which are all given in the same order as the '?' in the statement.
 *   When the statement is to be executed, concrete values must be passed
 *   on input. It is up to the user to pass values that are compatible with
 *   the types returned in the signature lists. (But it is guaranteed that
 *   at least the DBMS generates an error if the values are too incompatible.)
 *
 * - Transactions: Finishing a transaction (either committing or aborting it)
 *   closes all cursors. Statements remain prepared across transactions.
 *
 * - Errors: There are several sorts of errors:
 *   + No_rows_found: raised if there are no rows for this 'where' clause
 *     of the statement. This rather an indication of an important condition
 *     than an error. -- If you use streams to access your data, this exception
 *     is not raised. Instead, the stream ends.
 *   + Timeout_request: raised if another session has a lock that we are
 *     waiting for (too long). The statement is aborted.
 *   + Timeout_lock: raised if we have a lock too long. The statement is
 *     aborted.
 *   + Timeout_session: raised if the session takes too long. The session
 *     should be terminated. (This is not automatically done.)
 *   + Error: raised if the DBMS reports a special operating condition when 
 *      executing the statement. This aborts the execution of the statement.
 *   + Transaction_error: the DBMS reports an error that causes a rollback
 *     of the transaction. The rollback is automatically done.
 *   + Fatal_error: the DBMS reports an error that should terminate the
 *     session. (This is not automatically done.)
 *   + Failure: There are more error conditions detected by the driver itself
 *     that are simply reported as failures. These should normally be handled
 *     as Fatal_error.
 *
 * - Missing features:
 *   + support for insertion cursors. There should be a mechanism choosing
 *     the best available insertion method.
 *)

open Db_types

class virtual t =
  object (self)


    (***** Methods for opening and closing a connection *****)


    method virtual open_conn : Db_login.t -> unit

	(* Opens a new connection for this object. 
	 * REQUIRES: not self # state_is_open()
	 *)

    method virtual close_conn : unit -> unit
	
	(* Closes the current connection. The current transaction, if any,
	 * is rolled back.
	 * REQUIRES: self # state_is_open()
	 *)


    (***** Methods for transactions *****)


    method virtual commit : unit -> unit

        (* commits the current transaction.
	 * This implicitly releases all cursors opened in the transaction.
	 * REQUIRES: self # state_is_open()
	 *)

    method virtual rollback : unit -> unit
	
	(* rolls the current transaction back
	 * This implicitly releases all cursors opened in the transaction.
	 * REQUIRES: self # state_is_open()
	 *)


    (***** Methods for preparation and execution of simple statements *****)


    method virtual prepare : int -> string -> int

        (* prepares statements with given statement number. 
	 * returns the serial number of the prepared statement. As long as
	 * the serial number for a statement number is the same, the statement
	 * number has not been prepared again for a different statement.
	 * Normally, statements remain prepared if the current transaction
	 * is finishing and a new one begins.
	 *)

    method virtual describe : int -> (signature array * signature array)

	(* Returns the description of the statement prepared with the
	 * given statement number.
	 * The first component describes the input parameters and the second
	 * component the output parameters (host variables denoted by '?').
	 * There may be parameters acting both as input and output facility;
	 * these occur in both arrays.
	 * Note that description is not necessary for execution.
	 *)

    method virtual execute : int -> value array -> value array

        (* Executes the statement identified by the given number. The input
	 * parameters are filled with the value array passed on input,
	 * the output is returned by another value array.
	 * The order in the value arrays is the same as in the signature
	 * arrays returned by 'describe'.
	 * Behaviour is undefined if the number of input values does
	 * not match the number of input parameters, or if the type of a
	 * value is not compatible with the type of the corresponding 
	 * parameter.
	 *)


    (***** Statements with cursors *****)


    method virtual query : int -> value array -> value array

	(* Executes the given statement number (see 'execute') and declares
	 * and opens a cursor for it.
	 * The statement should be either a 'select' or a 'select for update'.
	 * The behaviour is undefined if the statement is of different type.
	 * Note that there are usually no output values.
	 *)

    method virtual describe_query : int -> signature array
	
	(* Describes the columns of the result table represented by the
	 * cursor for the given statement number.
	 * The cursor for the statement must already be open, i.e. the
	 * 'query' method has already been invoked.
	 *)

    method virtual fetch : int -> value array Stream.t

	(* Returns the results for the cursor for the given statement number
	 * as stream.
	 * The cursor is never automatically closed if the end of the stream
	 * is reached.
	 *)

    method virtual cursor_name : int -> string

	(* Returns the cursor name associated with the given statement.
	 * This name can be used to formulate an "update ... current of"
	 * statement for an update cursor.
	 *)

    method virtual release : int -> unit

	(* Closes the cursor for the given statement and releases its 
	 * resources.
	 *)


    (***** Methods for querying the state of the connection *****)


    method virtual state_is_open : unit -> bool
	
        (* returns true iff there is currently an open connection *)

    method virtual state_maximum_number_of_statements : unit -> int

	(* returns the maximum number of statements that can be prepared
	 * at the same time
	 *)

    method virtual state_serial_number : int -> int
	
	(* returns the serial number for a statement number. This number is
	 * negative if the statement number has never been used, and >= 0
	 * once a statement has been prepared using this number.
	 *)

    method virtual state_has_cursor : int -> bool

	(* returns true iff there is a cursor open for this statement
	 * number
	 *)

    method virtual state_next_stm_number : unit -> int

	(* suggests a statement number. It is guaranteed that this number
	 * does not have a cursor. 
	 * Returns Not_found if all statement numbers have cursors.
	 *)
end

(* ======================================================================
 * History:
 * 
 * $Log: dbll.ml,v $
 * Revision 1.1  1999/01/26 02:10:54  gerd
 * first release
 *
 * 
 *)
